from typing import Union, Type

from langchain.agents.output_parsers.openai_tools import OpenAIToolsAgentOutputParser
from langchain_core.agents import AgentFinish
from langchain_core.messages import AIMessage
from langchain_core.output_parsers import BaseOutputParser
from pydantic import BaseModel as BaseModelV2
from pydantic.v1 import BaseModel as BaseModelV1


class PydanticOutputParser(BaseOutputParser[Union[BaseModelV1, BaseModelV2]]):
    pydantic_class: Union[Type[BaseModelV1], Type[BaseModelV2]]
    base_parser: OpenAIToolsAgentOutputParser

    def parse_result(
            self, result: AIMessage, *, partial: bool = False
    ) -> Union[BaseModelV1, BaseModelV2]:
        tool_call_actions = self.base_parser.parse_result(result, partial=partial)  # noqa
        if isinstance(tool_call_actions, AgentFinish):
            raise ValueError("Only AgentAction with tool call may be converted to pydantic object")
        return self.pydantic_class(**tool_call_actions[0].tool_input)

    def parse(self, text: str) -> AgentFinish:
        raise ValueError("Can only parse messages")



import asyncio
import json
import tempfile
import uuid
from typing import Iterable, Union, Optional, Dict, List

import httpx
from openai import NotGiven, NOT_GIVEN, AsyncOpenAI
from openai._types import Headers, Query, Body
from openai._utils import maybe_transform, required_args
from openai.resources.chat import AsyncCompletions
from openai.types import ChatModel
from openai.types.chat import ChatCompletionMessageParam, completion_create_params, ChatCompletionStreamOptionsParam, \
    ChatCompletionToolChoiceOptionParam, ChatCompletionToolParam, ChatCompletion
from typing_extensions import Literal


class BatchesOpenAICompletions(AsyncCompletions):
    def __init__(self, client: AsyncOpenAI | None = None):
        super().__init__(client=client or AsyncOpenAI())

    def filter_not_given(self, params: Dict[str, str]) -> Dict[str, str]:
        return {key: value for key, value in params.items() if value != NOT_GIVEN}

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    async def create(
        self,
        *,
        messages: Iterable[ChatCompletionMessageParam],
        model: Union[str, ChatModel],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: Iterable[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        logprobs: Optional[bool] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        stream_options: Optional[ChatCompletionStreamOptionsParam] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: Iterable[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_logprobs: Optional[int] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # this params are ignored
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> ChatCompletion:
        job_id = str(uuid.uuid4())
        body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "logprobs": logprobs,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": False,
                "stream_options": NOT_GIVEN,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_logprobs": top_logprobs,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )
        request = {
            "body": self.filter_not_given(body),
            "custom_id": job_id,
            "method": "POST",
            "url": "/v1/chat/completions",
        }
        with tempfile.NamedTemporaryFile(suffix=".jsonl", mode="w") as input_file:
            # writing to input file
            input_file.write(json.dumps(request))
            input_file.flush()

            # uploading input file
            with open(input_file.name, "rb") as f:
                batch_input_file = await self._client.files.create(
                    file=f,
                    purpose="batch"
                )

        # launching
        batch_launch_info = await self._client.batches.create(
            input_file_id=batch_input_file.id,
            endpoint="/v1/chat/completions",
            completion_window="24h",
            metadata={"description": job_id}
        )

        # waiting to finish
        while True:
            await asyncio.sleep(5)
            batch_updated_info = await self._client.batches.retrieve(batch_launch_info.id)
            if batch_updated_info.status in ["validating", "in_progress", "finalizing", "cancelling"]:
                continue  # still busy
            elif batch_updated_info.status == "completed" and batch_updated_info.output_file_id:  # done
                # getting result
                result_content = await self._client.files.content(batch_updated_info.output_file_id)
                # parsing result
                json_string = result_content.content.decode(result_content.encoding)
                data_dict = json.loads(json_string)
                completion = ChatCompletion(**data_dict["response"]["body"])

                # cleanup files
                await self._client.files.delete(batch_input_file.id)
                await self._client.files.delete(batch_updated_info.output_file_id)

                return completion
            else:  # batch failed
                await self._client.files.delete(batch_input_file.id)
                if batch_updated_info.error_file_id:
                    result_content = await self._client.files.content(batch_updated_info.error_file_id)
                    json_string = result_content.content.decode(result_content.encoding)
                    data_dict = json.loads(json_string)
                    await self._client.files.delete(batch_updated_info.error_file_id)
                    raise Exception(data_dict["response"]["body"])
                else:
                    raise Exception("LLM failed with unknown reason")



from unittest import TestCase

from yid_langchain_extensions.output_parser.thoughts_json_parser import ThoughtsJSONParser
from yid_langchain_extensions.output_parser.utils import strip_json_from_md_snippet


class TestThoughtsJSONParser(TestCase):
    def setUp(self) -> None:
        self.json_md_snippet = """```json
{
    "thought1": "thought 1",
    "thought2": 2
}
```"""

    def test_strip_json_from_md_snippet(self):
        stripped_text = """
{
    "thought1": "thought 1",
    "thought2": 2
}
"""
        cleaned_text = strip_json_from_md_snippet(self.json_md_snippet)
        self.assertEqual(cleaned_text, stripped_text)

    def test_fix_json_md_snippet1(self):
        text = """```json
{
    "thought1": "thought 1",
    "thought2": 2"""
        fixed = ThoughtsJSONParser.fix_json_md_snippet(text)
        self.assertEqual(fixed, self.json_md_snippet)

    def test_fix_json_md_snippet2(self):
        text = """```json
{
    "thought1": "thought 1",
    "thought2": 2
}"""
        fixed = ThoughtsJSONParser.fix_json_md_snippet(text)
        self.assertEqual(fixed, self.json_md_snippet)

    def test_fix_json_md_snippet3(self):
        fixed = ThoughtsJSONParser.fix_json_md_snippet(self.json_md_snippet)
        self.assertEqual(fixed, self.json_md_snippet)

    def test_fix_json_md_snippet4(self):
        text = """
{
    "thought1": "thought 1",
    "thought2": 2"""
        fixed = ThoughtsJSONParser.fix_json_md_snippet(text)
        self.assertEqual(fixed, self.json_md_snippet)

    def test_fix_json_md_snippet5(self):
        text = """
{
    "thought1": "thought 1",
    "thought2": 2
}"""
        fixed = ThoughtsJSONParser.fix_json_md_snippet(text)
        self.assertEqual(fixed, self.json_md_snippet)

    def test_fix_json_md_snippet_extra_quotes(self):
        text = """"
{
    "thought1": "thought 1",
    "thought2": 2
}"""
        fixed = ThoughtsJSONParser.fix_json_md_snippet(text)
        self.assertEqual(fixed, self.json_md_snippet)



import unittest

import numpy as np
from langchain_core.messages import HumanMessage
from langchain_core.output_parsers import StrOutputParser
from langchain_openai import ChatOpenAI

from yid_langchain_extensions.llm.batches_openai_client import BatchesOpenAICompletions
from yid_langchain_extensions.utils import encode_image_to_url


class TestBatchesClient(unittest.IsolatedAsyncioTestCase):
    async def test_batches_client_in_chain(self):
        batches_client = BatchesOpenAICompletions()
        llm = ChatOpenAI(async_client=batches_client)
        chain = llm | StrOutputParser()
        answer = await chain.ainvoke(input=[HumanMessage(content="hi")])
        self.assertTrue(len(answer) > 0)

    def generate_random_image(self):
        random_image_array = np.random.randint(0, 256, (128, 128, 3), dtype=np.uint8)
        return random_image_array

    async def test_batches_client_with_inline_image(self):
        batches_client = BatchesOpenAICompletions()
        llm = ChatOpenAI(model_name="gpt-4o", async_client=batches_client)
        chain = llm | StrOutputParser()

        content = [
            {'type': 'text', 'text': 'hi'},
            {'type': 'image_url', 'image_url': {
                'url': encode_image_to_url(self.generate_random_image())
            }}
        ]

        answer = await chain.ainvoke(input=[HumanMessage(content=content)])
        self.assertTrue(len(answer) > 0)

    async def test_batches_client_with_online_image(self):
        batches_client = BatchesOpenAICompletions()
        llm = ChatOpenAI(model_name="gpt-4o", async_client=batches_client)
        chain = llm | StrOutputParser()

        url = ("https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/"
               "Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg")

        content = [
            {'type': 'text', 'text': 'hi'},
            {'type': 'image_url', 'image_url': {
                'url': url
            }}
        ]

        answer = await chain.ainvoke(input=[HumanMessage(content=content)])
        self.assertTrue(len(answer) > 0)



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='yid_langchain_extensions',
    version='0.4.48',
    author='Dmitrii Rashchenko',
    author_email='dimitree54@gmail.com',
    packages=find_packages(),
    description='Useful classes extending langchain library',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/dimitree54/yid_langchain_extensions',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=required
)



from abc import ABC
from typing import List, Optional

from langchain_core.output_parsers import BaseOutputParser


class StopSeqOutputParser(BaseOutputParser, ABC):
    @property
    def stop_sequences(self) -> Optional[List[str]]:
        return None



from typing import Any, Sequence, List, Tuple

from langchain.agents import Agent, AgentOutputParser
from langchain.base_language import BaseLanguageModel
from langchain.chains import LLMChain
from langchain.prompts import ChatPromptTemplate
from langchain.schema import AgentAction, BaseMessage, AIMessage, FunctionMessage
from langchain.schema.prompt_template import BasePromptTemplate
from langchain.tools import BaseTool

from yid_langchain_extensions.agent.raw_output_agent_executor import AgentWithThoughtsExecutor


class SimpleAgent(Agent):
    prompt: ChatPromptTemplate
    stop_sequences: List[str]

    @classmethod
    def from_llm_and_prompt(
            cls, llm: BaseLanguageModel, prompt: ChatPromptTemplate, output_parser: AgentOutputParser, **kwargs
    ) -> 'SimpleAgent':
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt
        )
        return cls(llm=llm, llm_chain=llm_chain, prompt=prompt, output_parser=output_parser, **kwargs)

    def _construct_scratchpad(
        self, intermediate_steps: List[Tuple[AgentAction, str]]
    ) -> List[BaseMessage]:
        """Construct the scratchpad that lets the agent continue its thought process."""
        thoughts: List[BaseMessage] = []
        for action, observation in intermediate_steps:
            thoughts.append(AIMessage(content=action.log))
            thoughts.append(FunctionMessage(name=action.tool, content=observation))
        return thoughts

    @property
    def _stop(self) -> List[str]:
        return self.stop_sequences

    @property
    def observation_prefix(self) -> str:
        return ""

    @property
    def llm_prefix(self) -> str:
        return ""

    @classmethod
    def create_prompt(cls, tools: Sequence[BaseTool]) -> BasePromptTemplate:
        raise NotImplementedError

    @classmethod
    def _get_default_output_parser(cls, **kwargs: Any) -> AgentOutputParser:
        raise NotImplementedError

    def get_executor(self, tools: List[BaseTool], **kwargs: Any) -> AgentWithThoughtsExecutor:
        return AgentWithThoughtsExecutor(agent=self, tools=tools, **kwargs)
